#!/usr/bin/env python3
#
# Check that the EI_EVENT_FOO and EIS_EVENT_FOO enum values match (for those events that
# exist in both libraries).

from typing import Iterator
from pathlib import Path
import re
import subprocess
import tempfile

template = """
#include <libei.h>
#include <libeis.h>

@@@

 int main(void) {
         return 0;
}
"""
assert_template = '_Static_assert(EIS_EVENT_{event} == EI_EVENT_{event}, "Mismatching event types for {event}");'


def extract(header: Path, prefix: str) -> Iterator[str]:
    with open(header) as fd:
        for line in fd:
            match = re.match(rf"^\t{prefix}(\w+)", line)
            if match:
                yield match[1]


if __name__ == "__main__":
    ei_events = extract(Path("src/libei.h"), prefix="EI_EVENT_")
    eis_events = extract(Path("src/libeis.h"), prefix="EIS_EVENT_")

    common = set(ei_events) & set(eis_events)

    print("Shared events that need identical values:")
    for e in common:
        print(f"  EI_EVENT_{e}")

    asserts = (assert_template.format(event=e) for e in common)

    with tempfile.NamedTemporaryFile(suffix=".c", delete=False) as fd:
        fd.write(template.replace("@@@", "\n".join(asserts)).encode("utf-8"))
        fd.flush()
        pkgconfig = subprocess.run(
            ["pkg-config", "--cflags", "--libs", "libei-1.0", "libeis-1.0"],
            check=True,
            capture_output=True,
        )
        pkgconfig_args = pkgconfig.stdout.decode("utf-8").strip().split(" ")
        try:
            subprocess.run(
                ["gcc", "-o", "event-type-check", *pkgconfig_args, fd.name],
                check=True,
                capture_output=True,
            )
            print("Success. Event types are identical")
        except subprocess.CalledProcessError as e:
            print("Mismatching event types")
            print(e.stdout.decode("utf-8"))
            print(e.stderr.decode("utf-8"))
            raise SystemExit(1)



# SPDX-License-Identifier: MIT
#
# This file is formatted with Python Black

from templates import Request, Response, Session, MockParams
from typing import Dict
from itertools import count

import dbus
import dbus.service
import enum
import logging

logger = logging.getLogger(f"templates.{__name__}")

BUS_NAME = "org.freedesktop.portal.Desktop"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.portal.RemoteDesktop"

_restore_tokens = count()


class RDSession(Session):
    class State(enum.IntEnum):
        CREATED = enum.auto()
        SELECTED = enum.auto()
        STARTED = enum.auto()
        CONNECTED = enum.auto()

    @property
    def state(self):
        try:
            return self._session_state
        except AttributeError:
            self._session_state = RDSession.State.CREATED
            return self._session_state

    def advance_state(self):
        if self.state != RDSession.State.CONNECTED:
            self._session_state += 1


def load(mock, parameters):
    logger.debug(f"loading {MAIN_IFACE} template")

    params = MockParams.get(mock, MAIN_IFACE)
    params.delay = 500
    params.version = parameters.get("version", 2)
    params.response = parameters.get("response", 0)
    params.devices = parameters.get("devices", 0b111)
    params.sessions: Dict[str, RDSession] = {}

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(params.version),
                "AvailableDeviceTypes": dbus.UInt32(
                    parameters.get("device-types", params.devices)
                ),
            }
        ),
    )


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="a{sv}",
    out_signature="o",
)
def CreateSession(self, options, sender):
    try:
        logger.debug(f"CreateSession: {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        session = RDSession(bus_name=self.bus_name, sender=sender, options=options)
        params.sessions[session.handle] = session

        response = Response(params.response, {"session_handle": session.handle})

        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="oa{sv}",
    out_signature="o",
)
def SelectDevices(self, session_handle, options, sender):
    try:
        logger.debug(f"SelectDevices: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        try:
            session = params.sessions[session_handle]
            if session.state != RDSession.State.CREATED:
                raise dbus.exceptions.DBusException(
                    f"Session in state {session.state}, expected CREATED",
                    name="org.freedesktop.DBus.Error.AccessDenied",
                )
            else:
                resp = params.response
                if resp == 0:
                    session.advance_state()
        except KeyError:
            raise dbus.exceptions.DBusException(
                "Invalid session", name="org.freedesktop.DBus.Error.AccessDenied"
            )

        response = Response(resp, {})
        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)
        if isinstance(e, dbus.exceptions.DBusException):
            raise e


@dbus.service.method(
    MAIN_IFACE,
    sender_keyword="sender",
    in_signature="osa{sv}",
    out_signature="o",
)
def Start(self, session_handle, parent_window, options, sender):
    try:
        logger.debug(f"Start: {session_handle} {options}")
        params = MockParams.get(self, MAIN_IFACE)
        request = Request(bus_name=self.bus_name, sender=sender, options=options)

        results = {
            "devices": dbus.UInt32(params.devices),
        }

        try:
            session = params.sessions[session_handle]
            if session.state != RDSession.State.SELECTED:
                raise dbus.exceptions.DBusException(
                    f"Session in state {session.state}, expected SELECTED",
                    name="org.freedesktop.DBus.Error.AccessDenied",
                )
            else:
                resp = params.response
                if resp == 0:
                    session.advance_state()
        except KeyError:
            raise dbus.exceptions.DBusException(
                "Invalid session", name="org.freedesktop.DBus.Error.AccessDenied"
            )

        response = Response(resp, results)
        request.respond(response, delay=params.delay)

        return request.handle
    except Exception as e:
        logger.critical(e)
        if isinstance(e, dbus.exceptions.DBusException):
            raise e


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}dd",
    out_signature="",
)
def NotifyPointerMotion(self, session_handle, options, dx, dy):
    try:
        logger.debug(f"NotifyPointerMotion: {session_handle} {options} {dx} {dy}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}udd",
    out_signature="",
)
def NotifyPointerMotionAbsolute(self, session_handle, options, stream, x, y):
    try:
        logger.debug(
            f"NotifyPointerMotionAbsolute: {session_handle} {options} {stream} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyPointerButton(self, session_handle, options, button, state):
    try:
        logger.debug(
            f"NotifyPointerButton: {session_handle} {options} {button} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}dd",
    out_signature="",
)
def NotifyPointerAxis(self, session_handle, options, dx, dy):
    try:
        logger.debug(f"NotifyPointerAxis: {session_handle} {options} {dx} {dx}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}ui",
    out_signature="",
)
def NotifyPointerAxisDiscrete(self, session_handle, options, axis, steps):
    try:
        logger.debug(
            f"NotifyPointerAxisDiscrete: {session_handle} {options} {axis} {steps}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyKeyboardKeycode(self, session_handle, options, keycode, state):
    try:
        logger.debug(
            f"NotifyKeyboardKeycode: {session_handle} {options} {keycode} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}iu",
    out_signature="",
)
def NotifyKeyboardKeysym(self, session_handle, options, keysym, state):
    try:
        logger.debug(
            f"NotifyKeyboardKeysym: {session_handle} {options} {keysym} {state}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}uudd",
    out_signature="",
)
def NotifyTouchDown(self, session_handle, options, stream, slot, x, y):
    try:
        logger.debug(
            f"NotifyTouchDown: {session_handle} {options} {stream} {slot} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}uudd",
    out_signature="",
)
def NotifyTouchMotion(self, session_handle, options, stream, slot, x, y):
    try:
        logger.debug(
            f"NotifyTouchMotion: {session_handle} {options} {stream} {slot} {x} {y}"
        )
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}u",
    out_signature="",
)
def NotifyTouchUp(self, session_handle, options, slot):
    try:
        logger.debug(f"NotifyTouchMotion: {session_handle} {options} {slot}")
    except Exception as e:
        logger.critical(e)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oa{sv}",
    out_signature="h",
)
def ConnectToEIS(self, session_handle, options):
    try:
        logger.debug(f"ConnectToEIS: {session_handle} {options}")

        params = MockParams.get(self, MAIN_IFACE)
        try:
            session = params.sessions[session_handle]
            if session.state != RDSession.State.STARTED:
                logger.error(f"Session in state {session.state}, expected STARTED")
                raise dbus.exceptions.DBusException(
                    "Session must be started before ConnectToEIS",
                    name="org.freedesktop.DBus.Error.AccesDenied",
                )
        except KeyError:
            raise dbus.exceptions.DBusException(
                "Invalid session", name="org.freedesktop.DBus.Error.AccessDenied"
            )

        import socket

        sockets = socket.socketpair()
        # Write some random data down so it'll break anything that actually
        # expects the socket to be a real EIS socket, plus it makes it
        # easy to check if we connected to the right EIS socket in our tests.
        sockets[0].send(b"VANILLA")
        fd = sockets[1]
        logger.debug(f"ConnectToEIS with fd {fd.fileno()}")
        return dbus.types.UnixFd(fd)
    except Exception as e:
        logger.critical(e)
        if isinstance(e, dbus.exceptions.DBusException):
            raise e



#!/usr/bin/env python3

# We can't import ei-scanner, so let's go via this route. The proper
# handling would be to have ei-scanner be the entry point for a ei_scanner
# module but.. well, one day we'll do that maybe.
import pytest

try:
    from eiscanner import parse, scanner, Protocol
except ImportError:
    print("Run tests from within the build directory")
    pytest.skip(allow_module_level=True)

from pathlib import Path

# set to the protocol file by meson
protofile = "@PROTOFILE@"


@pytest.fixture
def protocol_xml() -> Path:
    return Path(protofile)


@pytest.fixture
def protocol(protocol_xml: Path, component: str) -> Protocol:
    print(f"protocol for component {component}")
    return parse(protocol_xml, component)


@pytest.mark.skipif(
    protofile.startswith("@"),
    reason="Protocol XML file path invalid, run tests in the build dir",
)
class TestScanner:
    @pytest.mark.parametrize("component", ("eis", "ei", "brei"))
    def test_ei_names(self, component: str, protocol: Protocol):
        for interface in protocol.interfaces:
            assert interface.name.startswith(component)
            assert not interface.plainname.startswith(component)
            assert not interface.plainname.startswith("ei_")

        # just some manual checks
        assert "handshake" in [i.plainname for i in protocol.interfaces]
        assert "connection" in [i.plainname for i in protocol.interfaces]
        assert "button" in [i.plainname for i in protocol.interfaces]

    @pytest.mark.parametrize("component", ("ei",))
    def test_interface_arg(self, protocol: Protocol):
        intf = next((i for i in protocol.interfaces if i.name == "ei_device"))
        event = next((e for e in intf.events if e.name == "interface"))

        obj, interface_name, version = event.arguments
        assert obj.interface_arg == interface_name
        assert obj.interface_arg_for is None
        assert interface_name.interface_arg_for == obj
        assert interface_name.interface_arg is None
        assert version.interface_arg is None
        assert version.interface_arg_for is None

    def iterate_args(self, protocol: Protocol):
        for interface in protocol.interfaces:
            for request in interface.requests:
                for arg in request.arguments:
                    yield interface, request, arg
            for event in interface.events:
                for arg in event.arguments:
                    yield interface, event, arg

    @pytest.mark.parametrize("component", ("ei",))
    def test_versione_arg(self, protocol: Protocol):
        for interface, message, arg in self.iterate_args(protocol):
            if arg.protocol_type == "new_id":
                if f"{interface.plainname}.{message.name}" not in [
                    "connection.sync",
                ]:
                    assert (
                        arg.version_arg is not None
                    ), f"{interface.name}.{message.name}::{arg.name}"
                    assert (
                        arg.version_arg.name == "version"
                    ), f"{interface.name}.{message.name}::{arg.name}"
            elif arg.name == "version":
                if f"{interface.plainname}.{message.name}" not in [
                    "handshake.handshake_version",
                    "handshake.interface_version",
                ]:
                    assert (
                        arg.version_arg_for is not None
                    ), f"{interface.name}.{message.name}::{arg.name}"
                    assert (
                        arg.version_arg_for.name != "version"
                    ), f"{interface.name}.{message.name}::{arg.name}"
            else:
                assert (
                    arg.version_arg is None
                ), f"{interface.name}.{message.name}::{arg.name}"
                assert (
                    arg.version_arg_for is None
                ), f"{interface.name}.{message.name}::{arg.name}"

    @pytest.mark.parametrize("method", ("yamlfile", "jsonfile", "string"))
    def test_cli_extra_data(self, tmp_path, method):
        result_path = tmp_path / "result"
        tmpl_path = tmp_path / "template"
        with open(tmpl_path, "w") as template:
            template.write(">{{extra.foo}}<")

        if method == "yamlfile":
            extra_path = tmp_path / "extra_data.yml"
            with open(extra_path, "w") as extra_data:
                extra_data.write("foo: 'yes'")
            extra_data_arg = f"--jinja-extra-data-file={extra_path}"
        elif method == "jsonfile":
            extra_path = tmp_path / "extra_data.json"
            with open(extra_path, "w") as extra_data:
                extra_data.write('{"foo": "yes"}')
            extra_data_arg = f"--jinja-extra-data-file={extra_path}"
        elif method == "string":
            extra_data = '{"foo": "yes"}'
            extra_data_arg = f"--jinja-extra-data={extra_data}"
        else:
            pytest.fail(f"Unsupported method {method}")

        try:
            scanner(
                [
                    f"--output={result_path}",
                    extra_data_arg,
                    protofile,
                    str(tmpl_path),
                ]
            )
        except SystemExit as e:
            pytest.fail(reason=f"Unexpected system exit code {e}")

        assert result_path.exists()
        with open(result_path) as fd:
            result = fd.read()
            assert result == ">yes<"
